@@ -1,7 +1,6 @@
 from __future__ import unicode_literals
 
-from prompt_toolkit.completion import Completer, Completion
-from prompt_toolkit.contrib.completers import PathCompleter
+from prompt_toolkit.completion import Completer, Completion, PathCompleter
 from prompt_toolkit.contrib.regular_languages.compiler import compile as compile_grammar
 from prompt_toolkit.contrib.regular_languages.completion import GrammarCompleter
 
@@ -24,9 +23,32 @@ class PythonCompleter(Completer):
         self.get_globals = get_globals
         self.get_locals = get_locals
 
-        self._path_completer_grammar, self._path_completer = self._create_path_completer()
+        self._path_completer_cache = None
+        self._path_completer_grammar_cache = None
 
-    def _create_path_completer(self):
+    @property
+    def _path_completer(self):
+        if self._path_completer_cache is None:
+            self._path_completer_cache = GrammarCompleter(
+                self._path_completer_grammar, {
+                    'var1': PathCompleter(expanduser=True),
+                    'var2': PathCompleter(expanduser=True),
+                })
+        return self._path_completer_cache
+
+    @property
+    def _path_completer_grammar(self):
+        """
+        Return the grammar for matching paths inside strings inside Python
+        code.
+        """
+        # We make this lazy, because it delays startup time a little bit.
+        # This way, the grammar is build during the first completion.
+        if self._path_completer_grammar_cache is None:
+            self._path_completer_grammar_cache = self._create_path_completer_grammar()
+        return self._path_completer_grammar_cache
+
+    def _create_path_completer_grammar(self):
         def unwrapper(text):
             return re.sub(r'\\(.)', r'\1', text)
 
@@ -39,12 +61,20 @@ class PythonCompleter(Completer):
         grammar = r"""
                 # Text before the current string.
                 (
-                    [^'"#]            |  # Not quoted characters.
-                    '''.*'''          |  # Inside single quoted triple strings
-                    "" ".*"" "        |  # Inside double quoted triple strings
-                    \#[^\n]*          |  # Comment.
-                    "([^"\\]|\\.)*"   |  # Inside double quoted strings.
-                    '([^'\\]|\\.)*'      # Inside single quoted strings.
+                    [^'"#]                                  |  # Not quoted characters.
+                    '''  ([^'\\]|'(?!')|''(?!')|\\.])*  ''' |  # Inside single quoted triple strings
+                    "" " ([^"\\]|"(?!")|""(?!^)|\\.])* "" " |  # Inside double quoted triple strings
+
+                    \#[^\n]*(\n|$)           |  # Comment.
+                    "(?!"") ([^"\\]|\\.)*"   |  # Inside double quoted strings.
+                    '(?!'') ([^'\\]|\\.)*'      # Inside single quoted strings.
+
+                        # Warning: The negative lookahead in the above two
+                        #          statements is important. If we drop that,
+                        #          then the regex will try to interpret every
+                        #          triple quoted string also as a single quoted
+                        #          string, making this exponentially expensive to
+                        #          execute!
                 )*
                 # The current string that we're completing.
                 (
@@ -53,7 +83,7 @@ class PythonCompleter(Completer):
                 )
         """
 
-        g = compile_grammar(
+        return compile_grammar(
             grammar,
             escape_funcs={
                 'var1': single_quoted_wrapper,
@@ -63,10 +93,6 @@ class PythonCompleter(Completer):
                 'var1': unwrapper,
                 'var2': unwrapper,
             })
-        return g, GrammarCompleter(g, {
-            'var1': PathCompleter(),
-            'var2': PathCompleter(),
-        })
 
     def _complete_path_while_typing(self, document):
         char_before_cursor = document.char_before_cursor
@@ -88,7 +114,7 @@ class PythonCompleter(Completer):
                 yield c
 
         # If we are inside a string, Don't do Jedi completion.
-        if self._path_completer_grammar.match(document.text):
+        if self._path_completer_grammar.match(document.text_before_cursor):
             return
 
         # Do Jedi Python completions.
@@ -112,6 +138,22 @@ class PythonCompleter(Completer):
                 except ValueError:
                     # Jedi issue: "ValueError: invalid \x escape"
                     pass
+                except KeyError:
+                    # Jedi issue: "KeyError: u'a_lambda'."
+                    # https://github.com/jonathanslenders/ptpython/issues/89
+                    pass
+                except IOError:
+                    # Jedi issue: "IOError: No such file or directory."
+                    # https://github.com/jonathanslenders/ptpython/issues/71
+                    pass
+                except AssertionError:
+                    # In jedi.parser.__init__.py: 227, in remove_last_newline,
+                    # the assertion "newline.value.endswith('\n')" can fail.
+                    pass
+                except SystemError:
+                    # In jedi.api.helpers.py: 144, in get_stack_at_position
+                    # raise SystemError("This really shouldn't happen. There's a bug in Jedi.")
+                    pass
                 else:
                     for c in completions:
                         yield Completion(c.name_with_symbols, len(c.complete) - len(c.name_with_symbols),
